/* Copyright 2017, 2018 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/


/* Return the total POPCOUNT of a row of m many setwords.
   Same as setsize() (but without some 0s detection).
   */
int
nautyextra_row_popcount(const set *row, int m)
{
  int i, total = 0;
  for (i = 0; i < m; i++) {
    setword w = *row++;
    total += POPCOUNT(w);
  }
  return total;
}



void nautyextra_find_dist_sg (const sparsegraph *sg, int v, int *dist);

/* Set each dist[u] to the distance from v to u.
   Like find_dist() but for sparsegraph.

   Oops, same as distvals().
*/
void
nautyextra_find_dist_sg (const sparsegraph *sg, int v, int *dist)
{
  int            n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int from;
#if MAXN
  int queue[MAXN];
  assert(sg->nv <= MAXN);
#else
  DYNALLSTAT(int,queue,queue_sz);
  DYNALLOC1(int,queue,queue_sz,n,"nautyextra_find_dist_sg");
#endif

  if (n==0) return;
  for (from = 0; from < n; ++from) dist[from] = n;

  int *qhead = queue;
  int *qtail = queue;
  *qtail++ = v;
  dist[v] = 0;

  while (qhead != qtail) {
    assert (qhead < queue+n);
    int from = *qhead++;
    int epos = vv[from];
    int degree = dd[from];
    int to_i;
    for (to_i = 0; to_i < degree; to_i++) {
      int to = ee[epos+to_i];
      assert (0 <= to && to < n);
      int to_dist = dist[to];
      if (to_dist == n) {  /* not seen so far */
        dist[to] = dist[from] + 1;
        assert (qtail < queue+n);
        *qtail++ = to;
      }
    }
  }
}


int nautyextra_arc_count_sg (const sparsegraph *sg);

/* Return the number of arcs in sg.
   No, this is just sg->nde.
*/
int
nautyextra_arc_count_sg (const sparsegraph *sg)
{
  int n = sg->nv;
  const int *dd = sg->d;
  int count = 0;
  int v;
  for (v = 0; v < n; v++)
    count += *dd++;
  return count;
}





/* sg is an directed or undirected graph.
   Return the number of maximal independent sets in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   few vertices, and currently only uses 1-word sets, so must have sg->nv <=
   WORDSIZE many vertices.

   ENHANCE-ME: Could prune like nautyextra_indnum_and_count() below.
   ENHANCE-ME: Taking vertices in a neighbour-first order would then help
   prune more too.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_maximal_indsets_count_sg (const sparsegraph *sg)
{
  int n = sg->nv;
  setword count = 0;

  assert (n >= 0);
  if (n == 0) return 1;   /* n=0 empty set is only indset */
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_maximal_indsets_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  setword ss;
  NAUTYEXTRA_SET1_LOOP(ss,n,
                       count += nautyextra_is_indomset_sg_set(sg,&ss));
  return count;
}


#if 0
/* bits of indomnum_and_count */

      /* without = min(without_dom, without_undom)
         so without==without_undom means without_dom unknown
         or without <without_undom means without=without_dom
      */
      int                  calc_c_without_dom       = c_without;
      NAUTYEXTRA_TYPE_ULL  calc_c_without_dom_count = c_without_count;
      if (c_without_undom_count) {
        assert (c_without <= c_without_undom);
        if (c_without == c_without_undom) {
          assert (calc_c_without_dom_count >= c_without_undom_count);
          calc_c_without_dom_count -= c_without_undom_count;
        } else {
          calc_c_without_dom_count  = c_without_undom_count;
        }
      }
      DEBUG(printf("  without %d [%Lu]\n", c_without, c_without_count));
      DEBUG(printf("  calc without_dom %d [%Lu]\n",
                   calc_c_without_dom, calc_c_without_dom_count));
      assert(calc_c_without_dom       == c_without_dom);
      assert(calc_c_without_dom_count == c_without_dom_count);


      /* c_without = min(c_without_dom, c_without_undom)
         can have c_without_undom_count==0 for no sets of without_undom,
         eg. immediate parent of a leaf (the leaf must be present or that
         leaf would be undominated)
      */
      int                  calc_c_without       = c_without_dom;
      NAUTYEXTRA_TYPE_ULL  calc_c_without_count = c_without_dom_count;
      if (c_without_undom_count && c_without_undom < calc_c_without) {
        calc_c_without       = c_without_undom;
        calc_c_without_count = c_without_undom_count;
      } else if (c_without_undom == c_without) {
        ADD_NO_OVERFLOW_ULL(calc_c_without_count,
                            calc_c_without_count, c_without_undom_count);
      }
      DEBUG(printf("  c_without  %d [%Lu]\n", c_without, c_without_count));
      DEBUG(printf("  c_without  %d [%Lu] calc\n", calc_c_without, calc_c_without_count));
      assert(c_without_count != 0);
      assert(calc_c_without == c_without);
      assert(calc_c_without_count == c_without_count);


      /* v_without = sum c_dom,
         leave v_without variable before update */
      int                  v_without       = data[v].without;
      NAUTYEXTRA_TYPE_ULL  v_without_count = data[v].without_count;
      data[v].without = v_without + c_dom;
      MUL_NO_OVERFLOW_ULL(data[v].without_count,
                          v_without_count, c_dom_count);

#endif


/* ------------------------------------------------------------------------- */
/* Return the number of arcs u->v in sg.
   Oops, this is sg->nde.
*/
int nautyextra_arc_count_sg (const sparsegraph *sg, int u, int v);
int
nautyextra_arc_count_sg (const sparsegraph *sg, int u, int v)
{
  const int *ee = sg->e + sg->v[u];
  int degree    = sg->d[u];
  int count = 0;
  while (degree--) {
    int to = *ee++;
    count += (to == v);
  }
  return count;
}


/* ------------------------------------------------------------------------- */

/* Set *up,*vp to the vertices of a uniformly distributed random arc of g,m,n.
   If g has no arcs then set *up=*vp=-1.
*/
void
nautyextra_random_arc (const graph *g, int m, int n, int *up, int *vp)
{
  DEBUG(printf("nautyextra_random_arc()\n"));
#if MAXN
  int num_arcs[MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE(int,num_arcs,2*n,"nautyextra_random_arc");
#endif

  int u;
  int total_arcs = 0;
  for (u = 0; u < n; u++)
    total_arcs += (num_arcs[u] = setsize((set*)GRAPHROW(g,u,m), m));

  DEBUG(printf("  total_arcs %d\n", total_arcs));
  if (total_arcs == 0) {
    *up = *vp = -1;
    return;
  }
  int arc;
  GETKRAN(total_arcs,arc);
  DEBUG(printf("  arc %d\n", arc));

  for (u = 0; u < n; u++) {
    DEBUG(printf("u=%d arc %d\n", u, arc));
    assert (arc >= 0);

    if (arc < num_arcs[u]) {
      *up = u;
      *vp = nautyextra_set_nthelement(GRAPHROW(g,u,m), arc);
      assert (0 <= *up && *up < n);
      assert (0 <= *vp && *vp < n);
      return;
    }
    arc -= num_arcs[u];
  }
  DEBUG(printf("oops, end with arc still %d\n", arc));
  assert(0);
}

/* ------------------------------------------------------------------------- */
/* nautyextra_triangle_search() */

#if 0
      int c;
      for (c = b; (c = nextelement(a_row,m,c)) >= 0; ) {
        if (! ISELEMENT(b_row,c)) continue;  /* no b--c edge, not triangle */

        assert (a < b);
        assert (b < c);
        int ret = (*callback)(g,m,n, a,b,c, data);
        if (ret) return ret;
      }
#endif


/* ------------------------------------------------------------------------- */
/* nautyextra_make_random_vpar() */

#if 0
  DEBUG(printf("nautyextra_make_random_vpar() n=%d forest=%d\n", n, forest));
  assert (n >= 0);
  assert (forest==0 || forest==1);
  int limit = n + forest;
  int tree = 1 - forest;

  vpar[0] = 0;        /* seen */
  if (n) {
    int i;
    for (i = 1; i <= n; i++)
      vpar[i] = -1;     /* unseen */

    int r = 0;
    int upto = 1;
    int v = 0;
    for (i = 1; i < n; i++) {
      int next_r;
      GETKRAN(limit,next_r);
      next_r += tree;
      DEBUG(printf("at i=%d  r=%d next_r=%d\n", i, r, next_r));

      v = next_r;
      assert (0 <= v && v <= n);
      while (vpar[v] >= 0) {
        v = upto++;
        assert (1 <= v && v <= n);
      }

      DEBUG(printf("  store v=%d r=%d\n", v, r));
      assert (1 <= v && v <= n);
      vpar[v] = r;
      r = next_r;
    }

    do {
      v = upto++;
    } while (vpar[v] >= 0);
    DEBUG(printf("final store v=%d r=%d\n", v, r));
    assert (1 <= v && v <= n);
    vpar[v] = r;
  }
#endif

#if 0
  assert (n >= 0);
  assert (forest==0 || forest==1);
  if (n) {
    int i;
#if MAXN
    int  vec[MAXN];  /* 1 to n inclusive */
    assert(n <= MAXN);
#else
    DYNALLSTAT(int,vec,vec_sz);
    DYNALLOC1(int,vec,vec_sz,n,"nautyextra_make_random_vpar");
#endif

    int tree = 1 - forest;
    int limit = n+forest;
    for (i = 0; i < n-1; i++) {
      int r;
      GETKRAN(limit,r);
      vec[i] = r + tree;   /* 1 to n for tree, 0 to n for forest */
    }
    assert(i==n-1);
    vec[i] = 0;
    nautyextra_from_upascend(vec,n,0,vpar);
  }
#endif


/* Inferior version working downwards with aux data. */
void
nautyextra_depths_vpar (const int *vpar, int n, int *depths)
{
  DEBUG(printf("nautyextra_depths_vpar() n=%d\n", n));

  if (n==0) return;

  struct data {
    int first_child;
    int sibling;
    int queued;
  };
#if MAXN
  struct data  data[MAXN+1];    /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_depths_vpar");
#endif
  memset(data, '\0', (n+1)*sizeof(*data));

  int v;
  for (v = 1; v <= n; v++)
    depths[v] = -1;

  int queue_head = 0;
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    if (p) {
      int s = data[p].first_child;
      assert (1 <= p && p <= n);
      assert (0 <= s && s <= n);
      data[v].sibling = s;
      data[p].first_child = v;
    } else {
      assert (queue_head <= n);
      data[queue_head++].queued = v;   /* root queued */
    }
  }
  DEBUG(printf("queue_head %d\n", queue_head));

  int queue_tail = 0;
  while (queue_tail < queue_head) {
    int v = data[queue_tail++].queued;
    DEBUG(printf("queue take %d\n", v));

    int p = vpar[v];
    assert (0 <= p && p <= n);
    int depth = (p ? depths[p]+1 : 0);
    if (p) assert(depths[p] >= 0);
    do {
      DEBUG(printf(" set v=%d depth %d\n", v, depth));
      assert (depths[v] == -1);  /* not yet set */
      depths[v] = depth;

      int c = data[v].first_child;    /* queue child of v */
      if (c) {
        DEBUG(printf(" queue child c=%d\n", c));
        assert (queue_head <= n);
        data[queue_head++].queued = c;
      }
      v = data[v].sibling;   /* loop over siblings */
    } while (v);
  }
}
